/* _decode.c, by Robert Schmidt 1996 & W.H.Scholten 2001
 *
 * This file is released to the public domain
 */


#include "_decode_defs.h"

CircularBuffer<int> buffer(128);


int bytes_read = 0,
	dir = 0,
	dir_count = 0,
	bit_flank_sign = 1,
	bottom, top,
	funny_name;

const int min_flank_length = 5;

int  prev_found_flank = 0;

float average_flank = 0.0, threshold, bit_length, bit_error;
float low_0, high_0, low_1, high_1;


int next_sample()
{
#if 0 /* The original code */
	if (bufpos == maxbuflen) {
		buflen = read(audio_fd, inbuffer, maxbuflen);
		bufpos = 0;
	}

	bytes_read++;
	if (buflen < maxbuflen && bufpos == buflen)
		throw GiveUpInputEx("end of file", bytes_read);
#else /* The new code works much better for pcm input (fragsize reading into a ring buffer) */
	int  nbytes_read = 0;

	if (bufpos == maxbuflen) bufpos = 0;

	/* Fill up the internal buffer, read a 4K block */
	if (bufpos < bufend) {
		/* Add data at 'bufend' */
		int bytes_to_read = (min(4096, maxbuflen-bufend));
		if (bytes_to_read  > 0)
			nbytes_read = read(audio_fd, inbuffer+bufend, min(4096, maxbuflen-bufend));
	}
	else if (bufpos > bufend) {
		/* Do not read 1 byte at a time! */
		if (bufpos - bufend >= 4096) {
			nbytes_read = read(audio_fd, inbuffer+bufend, 4096);
		}
	} else {
		/* 2 possibilities: full or empty */
		if (buflen == 0) {
			bufpos = 0;
			nbytes_read = read(audio_fd, inbuffer, maxbuflen);
		}
	}

	if (nbytes_read < 0) {
		perror("read audio");
		exit(1);
	}

	buflen += nbytes_read;
	bufend += nbytes_read;
	if (bufend == maxbuflen) bufend = 0;

	/* Read a byte from the internal buffer */
	bytes_read++;
	buflen--;
	if (buflen == 0 && nbytes_read == 0)
		throw GiveUpInputEx("end of file", bytes_read);
#endif

	int b = inbuffer[ bufpos++ ] - 0x7f;
	return bit_flank_sign < 0 ? -b : b;
}


float find_flank(int dir)
{
	int *extreme = dir<0 ? &top : &bottom;
	int  dist;

	*extreme = buffer[0];
	for (;;) {
		buffer << next_sample();
		int delta = buffer[0]-*extreme;
		if (sign(delta) == -dir) {
			/* update extrema (top or bottom) */
			*extreme = buffer[0];
		}
		else {
			if (abs(delta) > (top-bottom)/3)
				/* We have moved from the last extrema by more than
				 * one third of the height between the last *two*
				 * extremas.  Then we have a flank. */
				break;
		}
	}
	dist = bytes_read - prev_found_flank;
	prev_found_flank = bytes_read;
	return dist;
}


void initial_synch()
{
	int i;
	int  dir = 1;
	const int flanks = 500;
	average_flank = 0.0;
	CircularBuffer<float> flank(flanks);

	if (debug) fprintf(stderr, "initial sync\n");

	/* skip first n flanks (but add them to averaging buffer) */
	for (i=0; i<flanks; i++) {
		flank << find_flank(dir); dir = -dir;
		average_flank += flank[0];
	}
	average_flank /= flanks;

	/* keep getting flanks until all flank lengths in buffer are close
	 * "enough" to the average */
	int good_flanks = 0;
	while (good_flanks < flanks) {
		average_flank -= flank[flanks-1] / flanks;
		flank << find_flank(dir);
		dir=-dir;
		average_flank += flank[0] / flanks;
		if (bound(flank[0], float(0.7*average_flank), float(1.3*average_flank)))
			good_flanks++;
		else
			good_flanks = 0;
	}

	if (debug) fprintf(stderr, "average flank distance %4.2f achieved at %d\n", average_flank, bytes_read);

	bit_length = 4*average_flank;

	if (debug) fprintf(stderr, "average bit length %4.2f\n", bit_length);

	bit_error = 0;

	low_1 = 0.3*average_flank;
	high_1 = 1.3*average_flank;
	low_0 = 1.3*average_flank;
	high_0 = 3.0*average_flank;

	if (debug) fprintf(stderr, "'0': %4.2f - %4.2f\n", low_0, high_0);
	if (debug) fprintf(stderr, "'1': %4.2f - %4.2f\n", low_1, high_1);

	if (debug) fprintf(stderr, "initial sync done\n");
}


int find_block_start_bit()
{
	for (;;) {
		bit_flank_sign = 1;
		int dir = 1;
		while (!bound(find_flank(dir), low_0, high_0)) {
			dir = -dir;  /* toggle expected flank direction */
		}
		if (!bound(find_flank(-dir), low_0, high_0)) {
			if (debug) fprintf(stderr, "bad start bit at %d\n", bytes_read);
			if (debug) fprintf(stderr, "top = %d, bottom =%d, buffer[0]=%d\n", top, bottom, buffer[0]);
			if (top-bottom < 10)
				throw GiveUpBlockEx("weak signal", bytes_read);
		}
		else {
			bit_flank_sign *= sign(buffer[0] - buffer[1]);
			if (debug) fprintf(stderr, "bit flank sign is %2d at %d\n", bit_flank_sign, bytes_read);
			return 1;
		}
	}
	return 0;
}


void find_wave()
{
	top = buffer[0];
	for (;;) {
		buffer << next_sample();
		if (buffer[0] > top)
			top = buffer[0];
		else
			if ((top - buffer[0]) >= (top - bottom)/2)
				break;
	}
	bottom = buffer[0];
	for (;;) {
		buffer << next_sample();
		if (buffer[0] < bottom)
			bottom = buffer[0];
		else
			if ((buffer[0] - bottom) >= (top - bottom)/2)
				break;
	}
}


int get_next_bit()
{
	int start = bytes_read, bit = -1;
	int prev_bytes_read = bytes_read;

	find_wave();

	if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
		bit = 0;
	else {
		prev_bytes_read = bytes_read;
		find_wave();
		if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
			bit = 1;
	}

	if (bit < 0) {
		if (debug) fprintf(stderr, "bad bit between %d and %d (%d, %d -> %4.2f, %4.2f)\n", start, bytes_read, prev_bytes_read - start, bytes_read - prev_bytes_read, (prev_bytes_read - start) / bit_length, (bytes_read - prev_bytes_read) / bit_length);
	}
 
	return bit;
}


int get_start_bit()
{
	int pos = bytes_read;

	if (get_next_bit() == 1) {
		if (debug) fprintf(stderr, "extra stop bit (?) at %d\n", pos);
		if (get_next_bit() != 0)
			throw GiveUpBlockEx("expected start bit (0)", pos);
	}

	return 1;
}


int get_stop_bit()
{
	int pos = bytes_read;

	if (get_next_bit() != 1)
		throw GiveUpBlockEx("expected stop bit (1)", pos);

	return 1;
}


int get_byte(int have_start_bit = 0)
{
	if (!have_start_bit)
		get_start_bit();

	int byte = 0;

	for (int bit=1; bit<=128; bit<<=1)
		if (get_next_bit()) /* Note: also a 1 for a bad bit ! */
			byte |= bit;

	get_stop_bit();

	return byte;
}


void get_block(struct bbctape_block *block)
{
	int  got_block = 0, byte, i, crc;
	int  hdr;

	while (!got_block) {
		try {
			memset(block->name, 0, 11);
			block->passed_crc = false;
			find_block_start_bit();

			byte = get_byte(1);
			switch (byte) {
			case 0xaa:
				if (debug) fprintf(stderr, "sync bug fix byte (&AA) found\n");
				break;

			case 0x2a:
				i=0;
				do {
					block->name[i++] = byte = get_byte();
				} while (byte != 0 && strlen(block->name) < 11);

				printf("Found : ");
				if (byte != 0) {
					printf("<filename too long>\n");
					throw GiveUpBlockEx("filename too long", bytes_read);
				}

				block->header[0] = '*';
				strcpy((char *) block->header+1, block->name);
				hdr = strlen((char *) block->header);

				for (i=0; i<CASSETTE_REMAINDER_HEADER_SIZE; i++) {
					block->header[ hdr + 1 + i ] = get_byte();
				}

				block->load_address = block->header [ hdr + 1 ] | (block->header [ hdr + 2 ] << 8) | (block->header [ hdr + 3 ] << 16) | (block->header [ hdr + 4 ] << 24);
				hdr += 4;

				block->exec_address = block->header [ hdr + 1 ] | (block->header [ hdr + 2 ] << 8) | (block->header [ hdr + 3 ] << 16) | (block->header [ hdr + 4 ] << 24);
				hdr += 4;

				block->number = block->header [ hdr + 1 ] | (block->header [ hdr + 2 ] << 8);
				hdr += 2;

				block->length = block->header [ hdr + 1 ] | (block->header [ hdr + 2 ] << 8);
				hdr += 2;

				block->flags = block->header [ hdr + 1 ];
				hdr += 1;

				/* 4 spare bytes */
				hdr += 4;

				block->header_crc = block->header [ hdr + 2 ] | (block->header [ hdr + 1 ] << 8); /* MSB first */


				block->header_length = CASSETTE_REMAINDER_HEADER_SIZE + 1 + strlen((char *) block->header);

				char printable_file[50];
				strcpy(printable_file, block->name);
				bbcim_printable_bbcfilename(printable_file);

				printf("%-10s %02X", printable_file, block->number);


				crc = bbcim_data_crc(block->header+1, CASSETTE_REMAINDER_HEADER_SIZE - 2 + strlen((char *) block->header+1)+1);
				if (crc != block->header_crc) {
					printf(" header CRC error (read %4X, calculated %4X)\n", block->header_crc, crc);
					/* We don't try to interpret any further data in case the header is corrupt, so: */
					break;
				}

				if (block->length) {
					unsigned char *ptr = (unsigned char *) realloc(block->data, block->length + 2);
					if (ptr) block->data = ptr;
					else { fprintf(stderr, "Out of memory\n"); exit(1); }

					for (i=0; i<block->length+2; i++) /* also the CRC */
						block->data[i] = get_byte();

//					block->data_crc = get_byte() << 8;  /* MSB first */
//					block->data_crc |= get_byte();
					block->data_crc = (block->data [ block->length ] << 8) | block->data [ block->length+1 ];  /* MSB first */

					crc = bbcim_data_crc(block->data, block->length);
					if (block->data_crc != crc) {
						printf(" data CRC error (read %4X, calculated %4X)", block->data_crc, crc);
						//break; /* commenting out enables bad block saving */
					}
					else block->passed_crc = true;
				} else block->passed_crc = true;

				printf("  (%8lX  %8lX  len %3X  flag %2X)", block->load_address, block->exec_address, block->length, block->flags);
				if (block->flags & 0x80) {
					printf(" Last block");
				}
				fflush(stdout);

				got_block = 1;
				break;

			default:
				if (debug) fprintf(stderr, "bad sync byte %2X\n", byte);
			}
		}
		catch (GiveUpBlockEx &e) {
//			if (strlen(block->name) > 0) {
//				printf("%s : Data?\n", block->name);
//			}
//			else {
				printf(" Data?\n");
//			}

			initial_synch();
		}

	}
}
